@@ -72,10 +72,10 @@ exception Diffie_hellman_error
 exception Ec_curve_error
 
 (** The SSL server certificate could not be initialized. *)
-exception Certificate_error
+exception Certificate_error of string
 
 (** The SSL server private key could not be intialized. *)
-exception Private_key_error
+exception Private_key_error of string
 
 (** The SSL private key does not match the certificate public key. *)
 exception Unmatching_keys
@@ -205,11 +205,12 @@ val get_error_string : unit -> string
 (** Protocol used by SSL. *)
 type protocol =
   | SSLv23 (** accept all possible protocols (SSLv2 if supported by openssl,
-               SSLv3, TLSv1, TLSv1.1 and TLSv1.2) *)
+               SSLv3, TLSv1, TLSv1.1, TLSv1.2, and TLSv1.3) *)
   | SSLv3 (** only SSL v3 protocol *)
   | TLSv1 (** only TLS v1 protocol *)
   | TLSv1_1 (** only TLS v1.1 protocol *)
   | TLSv1_2 (** only TLS v1.2 protocol *)
+  | TLSv1_3 (** only TLS v1.3 protocol *)
 
 (** An SSL abstract socket. *)
 type socket
@@ -275,6 +276,12 @@ val set_verify : context -> verify_mode list -> verify_callback option -> unit
 (** Set the maximum depth for the certificate chain verification that shall be allowed. *)
 val set_verify_depth : context -> int -> unit
 
+(** Set the list of supported ALPN protocols for negotiation to the context. *)
+val set_context_alpn_protos : context -> string list -> unit
+
+(** Set the callback to allow server to select the preferred protocol from client's available protocols. *)
+val set_context_alpn_select_callback : context -> (string list -> string option) -> unit
+
 
 (** {2 Ciphers} *)
 
@@ -347,6 +354,10 @@ val get_subject : certificate -> string
     given in arguments could not be found. *)
 val load_verify_locations : context -> string -> string -> unit
 
+(** Specifies that the default locations from which CA certificates are loaded
+    should be used. Returns [true] on success. *)
+val set_default_verify_paths : context -> bool
+
 (** Get the verification result. *)
 val get_verify_result : socket -> int
 
@@ -369,6 +380,12 @@ val shutdown_connection : socket -> unit
   * Name Indication (SNI) TLS extension. *)
 val set_client_SNI_hostname : socket -> string -> unit
 
+(** Set the list of supported ALPN protocols for negotiation to the connection. *)
+val set_alpn_protos : socket -> string list -> unit
+
+(** Get the negotiated protocol from the connection. *)
+val get_negotiated_alpn_protocol : socket -> string option
+
 (** Connect an SSL socket. *)
 val connect : socket -> unit
 
@@ -388,6 +405,20 @@ val shutdown : socket -> unit
     the peer, if any. Raises a [verify_error] on failure. *)
 val verify : socket -> unit
 
+(** Flags to specify how a certificate is matched against a given host name *)
+type x509_check_flag =
+  | Always_check_subject
+  | No_wildcards
+  | No_partial_wildcards
+  | Multi_label_wildcards
+  | Single_label_subdomains
+
+(* Specify how a certificate should be matched against the host name *)
+val set_hostflags : socket -> x509_check_flag list -> unit
+
+(* Set the expected host name to be verified. *)
+val set_host : socket -> string -> unit
+
 (** Get the file descriptor associated with a socket. It is primarly useful for
     [select]ing on it; you should not write or read on it. *)
 val file_descr_of_socket : socket -> Unix.file_descr
