@@ -48,6 +48,9 @@
 #include <openssl/err.h>
 #include <openssl/crypto.h>
 #include <openssl/tls1.h>
+#include <openssl/x509v3.h>
+
+#include "ocaml_ssl.h"
 
 #ifdef WIN32
 #include <windows.h>
@@ -102,6 +105,19 @@ static struct custom_operations socket_ops =
   custom_deserialize_default
 };
 
+/* Option types */
+
+#define Val_none Val_int(0)
+
+static value Val_some(value v)
+{
+  CAMLparam1(v);
+  CAMLlocal1(some);
+  some = caml_alloc(1, 0);
+  Store_field(some, 0, v);
+  CAMLreturn(some);
+}
+
 
 /******************
  * Initialization *
@@ -260,6 +276,12 @@ static int protocol_flags[] = {
     SSL_OP_NO_TLSv1_2
 #else
     0 /* not supported ,nothing to disable */
+#endif
+    ,
+#ifdef HAVE_TLS13
+    SSL_OP_NO_TLSv1_3
+#else
+    0 /* not supported, nothing to disable */
 #endif
 };
 
@@ -361,6 +383,25 @@ static const SSL_METHOD *get_method(int protocol, int type)
 #endif
       break;
 
+    case 5:
+#ifdef HAVE_TLS13
+      switch (type)
+      {
+        case 0:
+          method = TLS_client_method();
+          break;
+
+        case 1:
+          method = TLS_server_method();
+          break;
+
+        case 2:
+          method = TLS_method();
+          break;
+      }
+#endif
+      break;
+
     default:
       caml_leave_blocking_section();
       caml_invalid_argument("Unknown method (this should not have happened, please report).");
@@ -404,17 +445,20 @@ CAMLprim value ocaml_ssl_ctx_use_certificate(value context, value cert, value pr
   SSL_CTX *ctx = Ctx_val(context);
   char *cert_name = String_val(cert);
   char *privkey_name = String_val(privkey);
+  char buf[256];
 
   caml_enter_blocking_section();
   if (SSL_CTX_use_certificate_chain_file(ctx, cert_name) <= 0)
   {
+    ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
     caml_leave_blocking_section();
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string(buf));
   }
   if (SSL_CTX_use_PrivateKey_file(ctx, privkey_name, SSL_FILETYPE_PEM) <= 0)
   {
+    ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
     caml_leave_blocking_section();
-    caml_raise_constant(*caml_named_value("ssl_exn_private_key_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_private_key_error"), caml_copy_string(buf));
   }
   if (!SSL_CTX_check_private_key(ctx))
   {
@@ -520,6 +564,7 @@ CAMLprim value ocaml_ssl_ctx_set_client_CA_list_from_file(value context, value v
   SSL_CTX *ctx = Ctx_val(context);
   char *filename = String_val(vfilename);
   STACK_OF(X509_NAME) *cert_names;
+  char buf[256];
 
   caml_enter_blocking_section();
   cert_names = SSL_load_client_CA_file(filename);
@@ -527,14 +572,153 @@ CAMLprim value ocaml_ssl_ctx_set_client_CA_list_from_file(value context, value v
     SSL_CTX_set_client_CA_list(ctx, cert_names);
   else
   {
+    ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
     caml_leave_blocking_section();
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string(buf));
   }
   caml_leave_blocking_section();
 
   CAMLreturn(Val_unit);
 }
 
+#ifdef HAVE_ALPN
+static int get_alpn_buffer_length(value vprotos)
+{
+  value protos_tl = vprotos;
+  int total_len = 0;
+  while (protos_tl != Val_emptylist)
+  {
+    total_len += caml_string_length(Field(protos_tl, 0)) + 1;
+    protos_tl = Field(protos_tl, 1);
+  }
+  return total_len;
+}
+
+static void build_alpn_protocol_buffer(value vprotos, unsigned char *protos)
+{
+  int proto_idx = 0;
+  while (vprotos != Val_emptylist)
+  {
+    value head = Field(vprotos, 0);
+    int len = caml_string_length(head);
+    protos[proto_idx++] = len;
+
+    int i;
+    for (i = 0; i < len; i++)
+      protos[proto_idx++] = Byte_u(head, i);
+    vprotos = Field(vprotos, 1);
+  }
+}
+
+CAMLprim value ocaml_ssl_ctx_set_alpn_protos(value context, value vprotos)
+{
+  CAMLparam2(context, vprotos);
+  SSL_CTX *ctx = Ctx_val(context);
+
+  int total_len = get_alpn_buffer_length(vprotos);
+  unsigned char protos[total_len];
+  build_alpn_protocol_buffer(vprotos, protos);
+
+  caml_enter_blocking_section();
+  SSL_CTX_set_alpn_protos(ctx, protos, sizeof(protos));
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_unit);
+}
+
+static value build_alpn_protocol_list(const unsigned char *protocol_buffer, unsigned int len)
+{
+  CAMLparam0();
+  CAMLlocal3(protocol_list, current, tail);
+
+  int idx = 0;
+  protocol_list = Val_emptylist;
+
+  while (idx < len)
+  {
+    int proto_len = (int) protocol_buffer[idx++];
+    char proto[proto_len + 1];
+    int i;
+    for (i = 0; i < proto_len; i++)
+      proto[i] = (char) protocol_buffer[idx++];
+    proto[proto_len] = '\0';
+
+    tail = caml_alloc(2, 0);
+    Store_field(tail, 0, caml_copy_string(proto));
+    Store_field(tail, 1, Val_emptylist);
+
+    if (protocol_list == Val_emptylist)
+      protocol_list = tail;
+    else
+      Store_field(current, 1, tail);
+
+    current = tail;
+  }
+
+  CAMLreturn(protocol_list);
+}
+
+static int alpn_select_cb(SSL *ssl,
+                          const unsigned char **out,
+                          unsigned char *outlen,
+                          const unsigned char *in,
+                          unsigned int inlen,
+                          void *arg)
+{
+  CAMLparam0();
+  CAMLlocal3(protocol_list, selected_protocol, selected_protocol_opt);
+
+  int len;
+
+  caml_leave_blocking_section();
+  protocol_list = build_alpn_protocol_list(in, inlen);
+  selected_protocol_opt = caml_callback(*((value*)arg), protocol_list);
+
+  if (selected_protocol_opt == Val_none)
+    return SSL_TLSEXT_ERR_NOACK;
+
+  selected_protocol = Field(selected_protocol_opt, 0);
+  len = caml_string_length(selected_protocol);
+  *out = String_val(selected_protocol);
+  *outlen = len;
+  caml_enter_blocking_section();
+
+  return SSL_TLSEXT_ERR_OK;
+}
+
+CAMLprim value ocaml_ssl_ctx_set_alpn_select_callback(value context, value cb)
+{
+  CAMLparam2(context, cb);
+  SSL_CTX *ctx = Ctx_val(context);
+
+  value *select_cb;
+
+  select_cb = malloc(sizeof(value));
+  *select_cb = cb;
+  caml_register_global_root(select_cb);
+
+  caml_enter_blocking_section();
+  SSL_CTX_set_alpn_select_cb(ctx, alpn_select_cb, select_cb);
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_unit);
+}
+#else
+CAMLprim value ocaml_ssl_ctx_set_alpn_protos(value context, value vprotos)
+{
+  CAMLparam2(context, vprotos);
+  caml_raise_constant(*caml_named_value("ssl_exn_method_error"));
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_ssl_ctx_set_alpn_select_callback(value context, value cb)
+{
+  CAMLparam2(context, cb);
+  caml_raise_constant(*caml_named_value("ssl_exn_method_error"));
+  CAMLreturn(Val_unit);
+}
+#endif
+
 static int pem_passwd_cb(char *buf, int size, int rwflag, void *userdata)
 {
   value s;
@@ -762,16 +946,18 @@ CAMLprim value ocaml_ssl_read_certificate(value vfilename)
   char *filename = String_val(vfilename);
   X509 *cert = NULL;
   FILE *fh = NULL;
+  char buf[256];
 
   if((fh = fopen(filename, "r")) == NULL)
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string("couldn't open certificate file"));
 
   caml_enter_blocking_section();
   if((PEM_read_X509(fh, &cert, 0, 0)) == NULL)
   {
     fclose(fh);
+    ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
     caml_leave_blocking_section();
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string(buf));
   }
   fclose(fh);
   caml_leave_blocking_section();
@@ -787,16 +973,18 @@ CAMLprim value ocaml_ssl_write_certificate(value vfilename, value certificate)
   char *filename = String_val(vfilename);
   X509 *cert = Cert_val(certificate);
   FILE *fh = NULL;
+  char buf[256];
 
   if((fh = fopen(filename, "w")) == NULL)
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string("couldn't open certificate file"));
 
   caml_enter_blocking_section();
   if(PEM_write_X509(fh, cert) == 0)
   {
     fclose(fh);
+    ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
     caml_leave_blocking_section();
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string(buf));
   }
   fclose(fh);
   caml_leave_blocking_section();
@@ -808,13 +996,16 @@ CAMLprim value ocaml_ssl_get_certificate(value socket)
 {
   CAMLparam1(socket);
   SSL *ssl = SSL_val(socket);
+  char buf[256];
 
   caml_enter_blocking_section();
   X509 *cert = SSL_get_peer_certificate(ssl);
   caml_leave_blocking_section();
 
-  if (!cert)
-    caml_raise_constant(*caml_named_value("ssl_exn_certificate_error"));
+  if (!cert) {
+    ERR_error_string_n(ERR_get_error(), buf, sizeof(buf));
+    caml_raise_with_arg(*caml_named_value("ssl_exn_certificate_error"), caml_copy_string(buf));
+  }
 
   CAMLlocal1(block);
   block = caml_alloc_final(2, finalize_cert, 0, 1);
@@ -871,6 +1062,19 @@ CAMLprim value ocaml_ssl_ctx_load_verify_locations(value context, value ca_file,
   CAMLreturn(Val_unit);
 }
 
+CAMLprim value ocaml_ssl_ctx_set_default_verify_paths(value context)
+{
+  CAMLparam1(context);
+  int ret;
+  SSL_CTX *ctx = Ctx_val(context);
+
+  caml_enter_blocking_section();
+  ret = SSL_CTX_set_default_verify_paths(ctx);
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_bool(ret));
+}
+
 /*************************
  * Operations on sockets *
  *************************/
@@ -940,6 +1144,65 @@ CAMLprim value ocaml_ssl_set_client_SNI_hostname(value socket, value vhostname)
 }
 #endif
 
+#ifdef HAVE_ALPN
+CAMLprim value ocaml_ssl_set_alpn_protos(value socket, value vprotos)
+{
+  CAMLparam2(socket, vprotos);
+  SSL *ssl = SSL_val(socket);
+
+  int total_len = get_alpn_buffer_length(vprotos);
+  unsigned char protos[total_len];
+  build_alpn_protocol_buffer(vprotos, protos);
+
+  caml_enter_blocking_section();
+  SSL_set_alpn_protos(ssl, protos, sizeof(protos));
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_ssl_get_negotiated_alpn_protocol(value socket)
+{
+  CAMLparam1(socket);
+  SSL *ssl = SSL_val(socket);
+
+  const unsigned char *data;
+  unsigned int len;
+  SSL_get0_alpn_selected(ssl, &data, &len);
+
+  if (len == 0) CAMLreturn(Val_none);
+
+  /* Note: we use the implementation `caml_alloc_initialized_string` (which
+   * unfortunately requires OCaml >= 4.06) instead of `copy_string` here
+   * because the selected protocol in `data` is not NULL-terminated.
+   *
+   * From https://www.openssl.org/docs/man1.0.2/man3/SSL_get0_alpn_selected.html:
+   *   SSL_get0_alpn_selected() returns a pointer to the selected protocol in
+   *   data with length len. It is not NUL-terminated. data is set to NULL and
+   *   len is set to 0 if no protocol has been selected. data must not be
+   *   freed.
+   */
+  value proto = caml_alloc_string (len);
+  memcpy((char *)String_val(proto), (const char*)data, len);
+
+  CAMLreturn(Val_some(proto));
+}
+#else
+CAMLprim value ocaml_ssl_set_alpn_protos(value socket, value vprotos)
+{
+  CAMLparam2(socket, vprotos);
+  caml_raise_constant(*caml_named_value("ssl_exn_method_error"));
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_ssl_get_negotiated_alpn_protocol(value socket)
+{
+  CAMLparam1(socket);
+  caml_raise_constant(*caml_named_value("ssl_exn_method_error"));
+  CAMLreturn(Val_unit);
+}
+#endif
+
 CAMLprim value ocaml_ssl_connect(value socket)
 {
   CAMLparam1(socket);
@@ -977,6 +1240,57 @@ CAMLprim value ocaml_ssl_verify(value socket)
   CAMLreturn(Val_unit);
 }
 
+CAMLprim value ocaml_ssl_set_hostflags(value socket, value flag_lst)
+{
+  CAMLparam2(socket, flag_lst);
+  SSL *ssl = SSL_val(socket);
+  unsigned int flags = 0;
+
+  while (Is_block(flag_lst))
+  {
+    switch(Int_val(Field(flag_lst, 0)))
+    {
+      case 0:
+        flags |= X509_CHECK_FLAG_ALWAYS_CHECK_SUBJECT;
+        break;
+      case 1:
+        flags |= X509_CHECK_FLAG_NO_WILDCARDS;
+        break;
+      case 2:
+        flags |= X509_CHECK_FLAG_NO_PARTIAL_WILDCARDS;
+        break;
+      case 3:
+        flags |= X509_CHECK_FLAG_MULTI_LABEL_WILDCARDS;
+        break;
+      case 4:
+        flags |= X509_CHECK_FLAG_SINGLE_LABEL_SUBDOMAINS;
+        break;
+      default:
+        caml_invalid_argument("flags");
+    }
+    flag_lst = Field(flag_lst, 1);
+  }
+
+  caml_enter_blocking_section();
+  X509_VERIFY_PARAM_set_hostflags(SSL_get0_param(ssl), flags);
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_unit);
+}
+
+CAMLprim value ocaml_ssl_set1_host(value socket, value host)
+{
+  CAMLparam2(socket, host);
+  SSL *ssl = SSL_val(socket);
+  const char *hostname = String_val (host);
+
+  caml_enter_blocking_section();
+  X509_VERIFY_PARAM_set1_host (SSL_get0_param(ssl), hostname, 0);
+  caml_leave_blocking_section();
+
+  CAMLreturn(Val_unit);
+}
+
 CAMLprim value ocaml_ssl_write(value socket, value buffer, value start, value length)
 {
   CAMLparam2(socket, buffer);
@@ -1342,7 +1656,7 @@ static DH *load_dh_param(const char *dhfile)
 {
   DH *ret=NULL;
   BIO *bio;
-  
+
   if ((bio=BIO_new_file(dhfile,"r")) == NULL)
   	goto err;
   ret=PEM_read_bio_DHparams(bio,NULL,NULL,NULL);
